
/****************************************************************************
*
*	CREADOR: Juan Manuel Fernndez Muoz
*
***************************************************************************** 
*****************************************************************************
*
*	NOMBRE: fingerprint.c
*   
*
*	DESCRIPCIN:	Algoritmo de reconocimiento de huellas dactilares basado
*   				en correlacin de imgenes.
*
****************************************************************************/

// Ficheros de cabecera.
#include <csl.h>
#include <csl_mcbsp.h>
#include <stdio.h>
#include <math.h>

#include "dsk6713.h"
#include "fpc1010.h"
#include "spi.h"
#include "fingerprint.h"

// Funcin principal (main).
void main(void)
{
	unsigned char	*imageptr;		// Puntero para imagen del sensor.
	int n;							// Nmero del usuario.
	
	// Inicializacin de la placa.
	DSK6713_init();
	
	// Inicializacin del mdulo de leds.
    DSK6713_LED_init();
	
	// Iniciamos la funcionalidad de la librera CSL (Chip Support library).
	CSL_init();
	
	/* El McBSP0 es usado como interfaz SPI entre el DSP y el sensor FPC1010.
	Por defecto es usado para la comunicacin entre el DSP y el codec AIC23.
	Habilitamos el McBSP0 para el sensor con la siguiente funcin.*/
	DSK6713_rset(DSK6713_MISC , MCBSP1SEL);
	
	// Inicializamos la comunicacin SPI.
	SPI_init(2884615,0);
	
	imageptr = &image[0][0];	// Puntero al comienzo de la matriz donde almacenaremos la imagen del sensor.
	
	// Inicializando nmero de usuarios registrados a cero.
	for (n=0; n<NUMUSER; n++)
	{
		database[n].userenroll = 0;
	}
	
	// Generacin de tablas.
	gen_twiddle(tw, NC);
	copy_vector(tw, twi, NC);
	complex_conj_vector(twi, NC>>1);
	bit_rev(tw, NC>>1);
	bit_rev(twi, NC>>1);
	gen_phase_table(phasepolar , NF);
	
	// Apagamos los leds del sensor escribiendo 1 en DC_CNTL1 y DC_CNTL0.
	DSK6713_rset(DSK6713_DC_REG,3);
	
	printf("Inicializacin correcta.\n");
	
	// Bucle principal.
	waitForFp:
	
    while (!enrollFP && !verifyFP)
    {
		if(exitFP)
		{ 
       		printf ("Saliendo del programa.\n");
       		// Apagamos los leds del sensor escribiendo 1 en DC_CNTL1 y DC_CNTL0.
			DSK6713_rset(DSK6713_DC_REG,3);
       		exit();
       	}
	}
	    
    // Apagamos los leds del sensor escribiendo 1 en DC_CNTL1 y DC_CNTL0.
	DSK6713_rset(DSK6713_DC_REG,3);
    	
    // El sensor debe ser inicializado siempre antes de leer la imagen.
	sensorInit();
		
	// Leyendo imagen.
	printf ("Capturando imagen.\n");
	readImage(imageptr);
	printf ("Lectura terminada. Puede retirar el dedo.\n");
	
	if (verifyFP)
	{
		// Preprocesado de la imagen capturada.
		histogram (image, hist, NFi, NCi);
		preprocessing(image, hist, NFi, NCi);
		copy_image(image, imagecapture, NFi, NCi);
		
		// Peticin del nmero de usuario.
		waitNumUser1:
		printf("Introduzca su nmero de usuario: ");
		scanf("%d", &n);
		if (n<0 || n>=NUMUSER || database[n].userenroll==0)
		{
			printf("Nmero de usuario no vlido.\n");
			goto waitNumUser1;
		}
		copy_image(database[n].imageuser, imagenroll, NFi, NCi);
		
		// Clculo del desplazamiento.
		complex_image(imagecapture, fftcapture, NFi, NCi, NF, NC);
		fft2d(fftcapture, tw, tw, NF, NC);
		complex_image(imagenroll, fftenroll, NFi, NCi, NF, NC);
		fft2d(fftenroll, tw, tw, NF, NC);
		poc(fftcapture, fftenroll, NF, NC);
		fft2d(fftcapture, twi, twi, NF, NC);
		divide_ifft2d(fftcapture, NF, NC);
		high_peak_poc(fftcapture, coord, NF, NC);
		printf("Desplazamiento entre imgenes: %d filas, %d columnas.\n", coord[0], coord[1]);
		
		// Obtencin de la zona comn.
		desp_image(imagecapture, coord, 0, NFi, NCi);
		desp_image(imagenroll, coord, 1, NFi, NCi);
		
		// Transformacin a coordenadas polares.
		trans_polar_image(imagecapture, fftcapture, coord, phasepolar, NFi, NCi, NF, NC);
		trans_polar_image(imagenroll, fftenroll, coord, phasepolar, NFi, NCi, NF, NC);
		
		// Valor de similitud.
		fft2d(fftcapture, tw, tw, NF, NC);
		fft2d(fftenroll, tw, tw, NF, NC);
		poc(fftcapture, fftenroll, NF, NC);
		fft2d(fftcapture, twi, twi, NF, NC);
		divide_ifft2d(fftcapture, NF, NC);
		simil = add_max_poc(fftcapture, NF, NC);
		printf("Resultado de la comparacin = %f\n", simil);
		
		// Decisin.
		if (simil>=0.12)
		{
			printf("Identificacin satisfactoria.\n");
			// Encendemos el led verde escribiendo 0 DC_CNTL1 para indicar la coincidencia entre huellas.
			DSK6713_rset(DSK6713_DC_REG,1);
		}
		else
		{
			printf("Identificacin fallida.\n");
			// Encendemos el led rojo escribiendo 0 DC_CNTL0 para indicar la no coincidencia entre huellas.
			DSK6713_rset(DSK6713_DC_REG,2);
		}
		
		dummyVerify();
	}
	
	if (enrollFP)
	{
		// Preprocesado de la imagen capturada.
		histogram (image, hist, NFi, NCi);
		preprocessing(image, hist, NFi, NCi);
		
		// Peticin del nmero de usuario.
		waitNumUser2:
		printf("Introduzca el nmero de usuario elegido: ");
		scanf("%d", &n);
		if (n<0 || n>=NUMUSER)
		{
			printf("Nmero de usuario no vlido.\n");
			goto waitNumUser2;
		}
		database[n].userenroll = 1;
		copy_image(image, database[n].imageuser, NFi, NCi);
		
		printf("Registro satisfactorio.\n");
		
		dummyEnroll();
	}
	
    goto waitForFp;
}

//	Funcin para controlar el botn Registrar.
void dummyEnroll(void) 
{
    enrollFP=0;
}

//	Funcin para controlar el botn Verificar.
void dummyVerify(void)
{
    verifyFP=0;
}

/*	Funcin que genera una tabla con valores de cosenos y senos en el vector phase. Esta tabla ser usada por
	la funcin trans_polar_image y su objetivo es reducir el tiempo de ejecucin de dicha funcin.
	Parmetros de entrada:
	phase: vector de tamao 2*nf. En los elementos pares se almacenarn los valores de los cosenos, y en los
		   impares los correspondientes a los senos.
	nf: nmero de filas de la matriz que contendr el resultado de la funcin trans_polar_image.*/
void gen_phase_table(float* phase, int nf)
{
	float tempp, invp, incp;
	int p;
	
	tempp = 0;
	invp = 1.0/nf;
	incp = (PI/2)*invp;
	
	for (p=0; p<nf; p++)
	{
		phase[2*p] = cos(tempp);
		phase[2*p+1] = sin(tempp);
		
		tempp += incp;
	}
}

/*	Copia el contenido de la imagen imi en la imagen imo, ambas de tamao nfxnc.
	Parmetros de entrada:
		imi: imagen de entrada.
		imo: imagen que contendr la copia.
		nf: nmero de filas de la imagen a copiar.
		nc: nmero de columnas de la imagen a copiar.*/
void copy_image(unsigned char imi[][NCi], unsigned char imo[][NCi], int nf, int nc)
{
	int f, c;
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			imo[f][c] = imi[f][c];
		}
	}
}

/*	Funcin que calcula el histograma de la imagen im, el cual es almacenado en h.
	Parmetros de entrada:
		im: imagen de entrada.
		h: puntero al vector que contendr el histograma.
		nf: nmero de filas de la imagen.
		nc: nmero de columnas de la imagen.*/
void histogram(unsigned char im[][NCi], unsigned int* h, int nf, int nc)
{
	int f, c;
	
	for (f=0; f<256; f++)
	{
		h[f] = 0;
	}
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			h[im[f][c]]++;
		}
	}
}

/*	Funcin que aplica un preprocesado a la imagen im. El objetivo de ste es eliminar un poco de ruido,
	realzar las crestas (debido a que son las que aportan la informacin de la huella) y oscurecer los valles.
	Parmetros de entrada:
		im: imagen de entrada.
		h: puntero al vector que contiene el histograma de dicha imagen.
		nf: nmero de filas de la imagen.
		nc: nmero de columnas de la imagen.*/
void preprocessing(unsigned char im[][NCi], unsigned int* h, int nf, int nc)
{
	int f, c, start, final, i, end, min_hist;
	float b, m, x;
	
	i = 0;
	end = 0;
	min_hist = 10;
	
	while (end==0)
	{
		if (h[i]<min_hist)
		{
			i++;
		}
		else
		{
			start = i;
			end = 1;
		}
	}
	
	while (end==1)
	{
		if (h[i]>min_hist)
		{
			i++;
		}
		else
		{
			final = i-1;
			end = 0;
		}
	}
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			if (im[f][c]<start)
			{
				im[f][c] = start;
			}
			if (im[f][c]>final)
			{
				im[f][c] = final;
			}
		}
	}
		
	m =start-final;
	m = 255/m;
	b = final;
	b = -m*final;
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			x = im[f][c];
			im[f][c] = m*x + b;
		}
	}
}

/*	Esta funcin coge una ventana de tamao nfxnc del centro de la imagen im,
	la cual tiene un tamao nfixnci, y la pasa a la matriz compleja x para 
	posteriormente aplicarle la FFT 2D.
	Parmetros de entrada:
		im: imagen de entrada.
		x: matriz que contendr la informacin de la imagen en nmeros complejos.
		nfi: nmero de filas de im.
		nci: nmero de columnas de im.
		nf: nmero de filas de x.
		nc: nmero de columnas de x.*/
void complex_image(unsigned char im[][NCi], float x[][2*NC], int nfi, int nci, int nf, int nc)
{
	int f, c;
	int df, dc;
	
	df = (nfi-nf)/2;
	dc = (nci-nc)/2;
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			x[f][2*c] = im[f+df][c+dc];
			x[f][2*c+1] = 0;
		}
	}
}

/*	Funcin para generar la tabla twiddle.
	Parmetros de entrada:
		w: tabla donde se guardarn los valores twiddle necesarios para la FFT.
		n: nmero de elementos complejos a los que se les har la FFT.*/
void gen_twiddle(float* w, int n)
{
    double delta = 2 * PI / n;
    int i;
    
    for(i = 0; i < n/2; i++)
    {
        w[2 * i + 1] = sin(i * delta);
        w[2 * i] = cos(i * delta);
    }
}

/*	Funcin que copia un vector x en el vector y, ambos de tamao n.
	Parmetros de entrada:
		x: vector que contiene los elementos a copiar.
		y: vector que recibir los valores de x.
		n: nmero de elementos del vector x.*/
void copy_vector(float* x, float* y, int n)
{
	int i;
	
	for (i=0; i<n; i++)
	{
		y[i] = x[i];
	}
}

/*	Funcin que realiza el complejo conjugado a todos los elementos del vector x
	de n elementos complejos.
	Parmetros de entrada:
		x: vector complejo al que se le quiere hacer el complejo conjugado.
		n: nmero de elementos complejos del vector.*/
void complex_conj_vector(float* x, int n)
{
	int i;
	
	for (i=0; i<n; i++)
	{
		x[2*i+1] = -x[2*i+1];
	}
}

/*	Funcin que realiza un reordenamiento de los n elementos complejos del vector x
	mediante la operacin del bit inverso.
	Parmetros de entrada:
		x: vector al que se le har dicha transformacin.
		n: nmero de elementos complejos del vector.*/
void bit_rev(float* x, int n)
{
	int i, j, k;
	float rtemp, itemp;
	
	j = 0;
	
	for(i=1; i < (n-1); i++)
	{
		k = n >> 1;
		while(k <= j)
		{
			j -= k;
			k >>= 1;
		}
		j += k;
		if(i < j)
		{
			rtemp = x[j*2];
			x[j*2] = x[i*2];
			x[i*2] = rtemp;
			itemp = x[j*2+1];
			x[j*2+1] = x[i*2+1];
			x[i*2+1] = itemp;
		}
	}
}

/*	Dada la matriz compleja x, esta funcin realiza la traspuesta a la misma.	
	Parmetros de entrada:
		x: matriz compleja a la que se le har dicha transformacin.
		nf: nmero de filas.
		nc: nmero de columnas.*/
void trasp_matrix(float x[][2*NC], int nf, int nc)
{
	float tempr, tempi;
	int f, c, b;
	
	b = 0;
	
	for (f=0; f<nf; f++)
	{
		for (c=b; c<nc; c++)
		{
			tempr = x[f][2*c];
			tempi = x[f][2*c+1];
			x[f][2*c] = x[c][2*f];
			x[f][2*c+1] = x[c][2*f+1];
			x[c][2*f] = tempr;
			x[c][2*f+1] = tempi;
		}
		b++;
	}
}

/*	Funcin que realiza la FFT 2D a la matriz compleja x de tamao nfxnc.
	Parmetros de entrada:
		x: matriz a la que se le har la transformacin.
		tw_f: tabla twidle correspondiente a las filas.
		tw_c: tabla twidle correspondiente a las columnas.
		nf: nmero de filas.
		nc: nmero de columnas.*/
void fft2d(float x[][2*NC], float* tw_f, float* tw_c, int nf, int nc)
{
	int f, c;
	float* ptr;
	short nffft, ncfft;
	
	nffft = nf;
	ncfft = nc;
	for (f=0; f<nf; f++)
	{
		ptr = &x[f][0];
		DSPF_sp_cfftr2_dit(ptr, tw_f, ncfft);
		bit_rev(ptr, nc);
	}
	trasp_matrix(x, nf, nc);
	for (c=0; c<nc; c++)
	{
		ptr = &x[c][0];
		DSPF_sp_cfftr2_dit(ptr, tw_c, nffft);
		bit_rev(ptr, nf);
	}
	trasp_matrix(x, nf, nc);
}

/*	Funcin que realiza la divisin por el producto nfxnc (nmero de elementos) a todos los elementos
	de la matriz compleja x. Es aplicada despus de realizar una IFFT 2D.
	Parmetros de entrada:
		x: matriz compleja a la que se le har dicha transformacin.
		nf: nmero de filas.
		nc: nmero de columnas.*/
void divide_ifft2d(float x[][2*NC], int nf, int nc)
{
	int f, c;
	float inv;
	
	inv = nf*nc;
	inv = 1.0/inv;
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			x[f][2*c] = x[f][2*c]*inv;
			x[f][2*c+1] = x[f][2*c+1]*inv;
		}
	}
}

/*	Realiza la funcin POC (Phase Only Correlation, del ingls) al par de elementos indicados por f (fila)
	y c (columna) de las matrices complejas x e y. El resultado se guardar en el mismo elemento de
	la matriz x.
	Parmetros de entrada:
		x: matriz compleja 1.
		y: matriz compleja 2.
		f: fila del elemento al que se le aplica la funcin.
		c: columna del elemento al que se le aplica la funcin.*/
void poc_function(float x[][2*NC], float y[][2*NC], int f, int c)
{
	float tempr, tempi, tempabs;
	
	if ((x[f][2*c]==0) && (x[f][2*c+1]==0))
	{
		if ((y[f][2*c]==0) && (y[f][2*c+1]==0))
		{
			tempr = 1;
			tempi = 0;
		}
		else
		{
			tempr = 0;
			tempi = 0;
		}
	}
	else if ((y[f][2*c]==0) && (y[f][2*c+1]==0))
	{
		tempr = 0;
		tempi = 0;
	}
	else
	{
		tempr = x[f][2*c]*y[f][2*c] + x[f][2*c+1]*y[f][2*c+1];
		tempi = x[f][2*c+1]*y[f][2*c] - x[f][2*c]*y[f][2*c+1];
		tempabs = sqrt(tempr*tempr + tempi*tempi);
		tempr = tempr/tempabs;
		tempi = tempi/tempabs;
	}
	x[f][2*c] = tempr;
	x[f][2*c+1] = tempi;
}

/*	Funcin que realiza la funcin POC (Phase Only Correlation, del ingls) a las matrices complejas x e y,
	ambas de tamao nfxnc y previamente convertidas al dominio de Fourier, devolviendo el resultado en la
	matriz x. Para ello hace uso de la funcin poc_function y reduce el nmero de operaciones a realizar
	aprovechando la propiedad de simetra de la transformada de Fourier 2D.
	Parmetros de entrada:
		x: matriz compleja 1. En esta matriz se guardar el resultado de la funcin.
		y: matriz compleja 2.
		nf: nmero de filas.
		nc: nmero de columnas.*/
void poc(float x[][2*NC], float y[][2*NC], int nf, int nc)
{
	int f, c;
	
	poc_function(x, y, 0, 0);
	poc_function(x, y, 0, nc/2);
	poc_function(x, y, nf/2, 0);
	poc_function(x, y, nf/2, nc/2);
	
	for (f=1; f<nf/2; f++)
	{
		c=0;
		poc_function(x, y, f, c);
		x[nf-f][0] = x[f][2*c];
		x[nf-f][1] = -x[f][2*c+1];
		c=nc/2;
		poc_function(x, y, f, c);
		x[nf-f][nc] = x[f][2*c];
		x[nf-f][nc+1] = -x[f][2*c+1];
	}
	
	for (c=1; c<nc/2; c++)
	{
		f=0;
		poc_function(x, y, f, c);
		x[0][2*nc-2*c] = x[f][2*c];
		x[0][2*nc-2*c+1] = -x[f][2*c+1];
		f=nf/2;
		poc_function(x, y, f, c);
		x[nf/2][2*nc-2*c] = x[f][2*c];
		x[nf/2][2*nc-2*c+1] = -x[f][2*c+1];
	}
	
	for (f=1; f<nf/2; f++)
	{
		for (c=1; c<nc/2; c++)
		{
			poc_function(x, y, f, c);
			x[nf-f][2*nc-2*c] = x[f][2*c];
			x[nf-f][2*nc-2*c+1] = -x[f][2*c+1];
		}
	}
	
	for (f=nf/2+1; f<nf; f++)
	{
		for (c=1; c<nc/2; c++)
		{
			poc_function(x, y, f, c);
			x[nf-f][2*nc-2*c] = x[f][2*c];
			x[nf-f][2*nc-2*c+1] = -x[f][2*c+1];
		}
	}
}

/*	Funcin que encuentra los dos valores ms altos dados por la aplicacin de la funcin POC a dos matrices
	complejas y devuelve su suma.
	Parmetros de entrada:
		x: matriz compleja con el resultado de la funcin POC a la que se le ha hecho la IFFT 2D
		   previamente para convertirla al dominio espacial.
		nf: nmero de filas.
		nc: nmero de columnas.*/
float add_max_poc(float x[][2*NC], int nf, int nc)
{
	int f, c;
	float max1, max2;
	
	max1 = 0;
	max2 = 0;
	
	for (f=0; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			if (x[f][2*c] > max1)
			{
				max1 = x[f][2*c];
			}
			else if (x[f][2*c] > max2) 
			{
				max2 = x[f][2*c];
			}
		}
	}
	max1 += max2;
	return max1;
}

/*	Funcin que encuentra la posicin del pico ms alto devuelto por la funcin POC.
	Parmetros de entrada:	
		x: matriz compleja con el resultado de la funcin POC a la que se le ha hecho la IFFT 2D previamente.
		nf: nmero de filas.
		nc: nmero de columnas.*/
void high_peak_poc(float x[][2*NC], int* coor, int nf, int nc)
{
	int f, c, i, j;
	float max;
	
	i = 0;
	j = 0;
	max = 0;
	
	for (f=0; f<nf; f++)
	{
		for (c=1; c<nc; c++)
		{
			if (x[f][2*c]>max)
			{
				max = x[f][2*c];
				i = f;
				j = c;
			}
		}
	}
	
	if (x[0][0]/10>max)
	{
		i = 0;
		j = 0;
	}
	else
	{
		if (i>nf/2)
		{
			i = -(nf-i);
		}
		else
		{
			i = i;
		}
		if (j>nc/2)
		{
			j = nc-j;
		}
		else
		{
			j = -j;
		}
	}
	coor[0] = i;
	coor[1] = j;
}

/*	Funcin que recorta la imagen im segn el desplazamiento calculado previamente respecto a otra imagen
	indicado por el vector coor. Con esto se extrae la zona comn de ambas imgenes.
	Parmetros de entrada:
		im: imagen que ser recortada.
		coor: desplazamiento entre huellas.
		enr: = 0 => se refiere a la imagen capturada; = 1 => se refiere a la imagen de la base de datos.
		nf: nmero de filas de las imgenes.
		nc: nmero de columnas de las imgenes.*/
void desp_image(unsigned char im[][NCi], int* coor, int enr, int nf, int nc)
{
	int f, c, df, dc, absdf, absdc;
	
	if (enr==1)
	{	
		if (coor[0]>=0 && coor[1]>=0)
		{
			df = 0;
			dc = coor[1];
			absdf = coor[0];
			absdc = coor[1];
		}
		else if (coor[0]>=0 && coor[1]<0)
		{
			df = 0;
			dc = 0;
			absdf = coor[0];
			absdc = -coor[1];
		}
		else if (coor[0]<0 && coor[1]>=0)
		{
			df = -coor[0];
			dc = coor[1];
			absdf = -coor[0];
			absdc = coor[1];
		}
		else
		{
			df = -coor[0];
			dc = 0;
			absdf = -coor[0];
			absdc = -coor[1];
		}
	}
	else
	{
		if (coor[0]>=0 && coor[1]>=0)
		{
			df = coor[0];
			dc = 0;
			absdf = coor[0];
			absdc = coor[1];
		}
		else if (coor[0]>=0 && coor[1]<0)
		{
			df = coor[0];
			dc = -coor[1];
			absdf = coor[0];
			absdc = -coor[1];
		}
		else if (coor[0]<0 && coor[1]>=0)
		{
			df = 0;
			dc = 0;
			absdf = -coor[0];
			absdc = coor[1];
		}
		else
		{
			df = 0;
			dc = -coor[1];
			absdf = -coor[0];
			absdc = -coor[1];
		}
	}
	
	for (f=0; f<nf-absdf; f++)
	{
		for (c=0; c<nc-absdc; c++)
		{
			im[f][c] = im[f+df][c+dc];
		}
		for (c=nc-absdc; c<nc; c++)
		{
			im[f][c] = 0;
		}
	}
	for (f=nf-absdf; f<nf; f++)
	{
		for (c=0; c<nc; c++)
		{
			im[f][c] = 0;
		}
	}
}

/*	Funcin para pasar a la matriz compleja x la imagen im en coordenadas polares atendiendo a su desplazamiento.
	Parmetros de entrada:
		im: imagen de entrada.
		x: matriz compleja a la que se le pasar la informacin (filas => fase; columnas => radio).
		coor: desplazamiento entre imgenes.
		phase: dicho vector contiene valores de senos y cosenos usados en esta funcin para reducir su tiempo
			de ejecucin.
		nfi: nmero de filas de im.
		nci: nmero de columnas de im.
		nf: nmero de filas de x.
		nc: nmero de columnas de x.*/
void trans_polar_image(unsigned char im[][NCi], float x[][2*NC], int* coor, float* phase, int nfi, int nci, int nf, int nc)
{
	int m, p, f, c, absdf, absdc, mmax;
	float tempf, tempc, tempm;
	float invm, incm;
	
	tempm=0;
	
	if (coor[0]>=0)
	{
		absdf = coor[0];
	}
	else
	{
		absdf = -coor[0];
	}
	if (coor[1]>=0)
	{
		absdc = coor[1];
	}
	else
	{
		absdc = -coor[1];
	}
	
	if ((nfi-absdf)>(nci-absdc))
	{
		mmax = nci-absdc;
	}
	else
	{
		mmax = nfi-absdf;
	}
	
	invm = 1.0/nc;
	incm = mmax*invm;
		
	for (m=0; m<nc; m++)
	{
		for (p=0; p<nf; p++)
		{
			tempf = tempm*phase[2*p+1];
			tempc = tempm*phase[2*p];
			f = tempf;
			c = tempc;
			
			x[p][2*m] = im[f][c];
			x[p][2*m+1] = 0;
		}
		
		tempm += incm;
	}
}
